/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.engine.cache;

import java.io.Serializable;
import java.security.Principal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.logging.*;

import com.funambol.framework.core.*;
import com.funambol.framework.engine.*;
import com.funambol.framework.engine.source.*;
import com.funambol.framework.security.*;
import com.funambol.framework.tools.beans.*;

import com.funambol.email.admin.EmailConnectorConfig;
import com.funambol.email.exception.EmailConfigException;
import com.funambol.email.util.Def;


/**
 * This class represents a SyncSource that read and store the state of the items
 * in a support table.
 *
 * @version $Id: CrcCachedSyncSource.java,v 1.11 2006/04/03 16:40:30 gbmiglia Exp $
 *
 */
public abstract class CrcCachedSyncSource
        extends AbstractSyncSource
        implements FilterableSyncSource, Serializable, LazyInitBean {

    // --------------------------------------------------------------- Constants

    private Logger log = Logger.getLogger(Def.LOGGER_NAME);

    public String jndiDataSourceName = null;

    // ------------------------------------------------------------ Private Data

    /**    */
    private CrcItemStateDAO crcItemStateDao = null;

    /**    */
    private Principal principal          = null;
    /**    */
    private int       syncMode;

    /**
     *
     */
    private CrcSyncItemInfo[] localItems    = null;
    private CrcSyncItemInfo[] serverItems   = null;

    /*
     *
     */
    private ArrayList newItemKeys     = null;
    private ArrayList deletedItemKeys = null;
    private ArrayList updatedItemKeys = null;

    /**
     * keep the items id and the result of send operation
     *
     */
    public ArrayList itemsSent = new ArrayList();



    // ---------------------------------------------------------- Public Methods

    /**
     * Init the syncSource
     *
     * @throws BeanInitializationException
     */
    public void init() throws BeanInitializationException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Init CrcCachedSyncSource with sourceUri: " + sourceURI);
        }

        try {
          jndiDataSourceName  = EmailConnectorConfig.getConfigInstance().getDataSource();
        } catch (EmailConfigException e) {
            log.severe("Error Getting jndi datasource Parameters:: " + e);
            log.throwing(getClass().getName(), "checkMailCredentials", e);
        }

        crcItemStateDao = new CrcItemStateDAO(jndiDataSourceName);
        try {
            crcItemStateDao.init();
        } catch (Exception e) {
            throw new BeanInitializationException("Error initializing CrcItemStateDAO", e);
        }
    }

    /**
     * BeginSync
     *
     * @param context The sync context that includes the principal,
     * sync mode and filter clause.
     * @throws SyncSourceException
     */
    public void beginSync(SyncContext context)
        throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource beginSync [" + sourceURI + "] ");
        }

        this.principal = context.getPrincipal();
        this.syncMode  = context.getSyncMode();

        long principalId = ((Sync4jPrincipal) this.principal).getId();

        try {


            if (syncMode == AlertCode.SLOW){
                 
                crcItemStateDao.deleteLocalItems(this.sourceURI,principalId);
                crcItemStateDao.deleteSentPop(this.sourceURI,principalId);

            } else if (syncMode == AlertCode.REFRESH_FROM_SERVER){
                
                crcItemStateDao.deleteLocalItems(this.sourceURI,principalId);
                crcItemStateDao.deleteSentPop(this.sourceURI,principalId);

            } else if (syncMode == AlertCode.REFRESH_FROM_CLIENT) {
                
                // delete all info on the server
                crcItemStateDao.deleteLocalItems(this.sourceURI,principalId);
                crcItemStateDao.deleteSentPop(this.sourceURI,principalId);
                clearServerInfo();
                
            } else {
                
                this.serverItems = getAllSyncItemInfo();
                
                this.localItems  = crcItemStateDao.getLocalItems(this.sourceURI,
                                                                 principalId);

                // create the array with the new, updated, deleted items
                this.newItemKeys     = CrcCachedSyncSourceHelper.setNewItemKeys(
                                              this.localItems, this.serverItems);
                this.updatedItemKeys = CrcCachedSyncSourceHelper.setUpdatedItemKeys(
                                              this.localItems, this.serverItems);
                this.deletedItemKeys = CrcCachedSyncSourceHelper.setDeletedItemKeys(
                                              this.localItems, this.serverItems);
            }


        } catch (Exception e) {
            throw new SyncSourceException("Error accessing items: " +
                                           e.getMessage(), e);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Crc Cached NEW");
            if (this.newItemKeys != null){
               log.finest("Crc Cached NEW num:" + this.newItemKeys.size());
            } else {
               log.finest("Crc Cached NO NEW ITEMS");
            }

            log.finest("Crc Cached UPDATED");
            if (this.updatedItemKeys != null){
               log.finest("Crc Cached UPDATED num:" + this.updatedItemKeys.size());
            } else {
               log.finest("Crc Cached NO UPDATED ITEMS");
            }

            log.finest("Crc Cached DELETED");
            if (this.deletedItemKeys != null){
               log.finest("Crc Cached DELETED num:" + this.deletedItemKeys.size());
            } else {
               log.finest("Crc Cached NO DELETED ITEMS");
            }
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource beginSync done [" + sourceURI + "]");
        }

    }

    /**
     * EndSync Method
     *
     * @todo optimize the performance, by read only update / add items
     * to setting lastModified with addItems[exist], deleteItems[exist],
     * updateItems[implements] array
     *
     *
     * @throws SyncSourceException
     */
    public void endSync() throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource endSync [" + sourceURI + "]"
                       + " principal " + this.principal );
        }

        CrcSyncItemInfo[] serverItems = null;

        long principalId = ((Sync4jPrincipal) this.principal).getId();

        // reload cache
        try {
            serverItems  = getAllSyncItemInfo();

            if (serverItems != null){
               if (log.isLoggable(Level.FINEST)) {
                  log.finest("CrcCachedSyncSource server Items num: " + serverItems.length);
               }
            }
            crcItemStateDao.updateLocalItems (serverItems ,
                                              this.sourceURI ,
                                              principalId);

        } catch (Exception e) {
            throw new SyncSourceException("Error caching items:" +
                                          e.getMessage(), e);
        }



        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource endSync done [" + sourceURI + "]");
        }

    }

    /**
     *
     * @see SyncSource
     *
     * @param since Timestamp
     * @param to Timestamp
     * @return SyncItemKey[]
     * @throws SyncSourceException
     */
    public SyncItemKey[] getUpdatedSyncItemKeys(Timestamp since, Timestamp to)
            throws SyncSourceException {
        return (SyncItemKey[])updatedItemKeys.toArray(new SyncItemKey[0]) ;
    }

    /**
     *
     * @see SyncSource
     *
     * @param since Timestamp
     * @param to Timestamp
     * @return SyncItemKey[]
     * @throws SyncSourceException
     */
    public SyncItemKey[] getNewSyncItemKeys(Timestamp since, Timestamp to)
            throws SyncSourceException {
        return (SyncItemKey[])newItemKeys.toArray(new SyncItemKey[0]);
    }

    /**
     *
     * @see SyncSource
     *
     * @param since Timestamp
     * @param to Timestamp
     * @return SyncItemKey[]
     * @throws SyncSourceException
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Timestamp since, Timestamp to)
            throws SyncSourceException {
        return (SyncItemKey[])deletedItemKeys.toArray(new SyncItemKey[0]) ;
    }



    /**
     *
     * @return CrcSyncItemInfo[]
     * @throws SyncSourceException
     */
    public abstract CrcSyncItemInfo[] getAllSyncItemInfo()
            throws  SyncSourceException;


    /**
     *
     * @throws SyncSourceException
     */
    public abstract void clearServerInfo()
            throws  SyncSourceException;

    
    /**
     *
     *
     * @param syncItemKey SyncItemKey
     * @return char
     * @throws SyncSourceException
     */
    public char getSyncItemStateFromId(SyncItemKey syncItemKey)
      throws SyncSourceException {

        char status = SyncItemState.UNKNOWN;
        
        status = CrcCachedSyncSourceHelper.getSyncItemStateFromId(syncItemKey,
                                                          this.syncMode,
                                                          this.newItemKeys, 
                                                          this.deletedItemKeys,
                                                          this.updatedItemKeys);
        
        return status;

    }

    /**
     *
     *
     * @param id String
     * @throws SyncSourceException
     */
    public void insertSentItem(String id) throws SyncSourceException {

        long principalId = ((Sync4jPrincipal) this.principal).getId();
        
        CrcCachedSyncSourceHelper.insertSentItem(id, 
                                                 this.sourceURI,  
                                                 principalId,
                                                 this.crcItemStateDao);
        
    }

}
