/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.items.manager;


import com.funambol.email.exception.EntityException;
import com.funambol.email.items.model.ItemMessage;
import com.funambol.email.items.model.InternalPart;
import com.funambol.email.util.Def;
import com.funambol.email.util.Utility;

import javax.mail.*;
import javax.mail.internet.*;
import javax.activation.DataHandler;
import java.io.IOException;
import java.util.*;
import java.util.logging.Level;


/**
 * <p>Filter class. </p>
 */

public class EntityManagerFilter extends EntityManager {
    
    //---------------------------------------------------------------- CONSTANTS
    
    /**        */
    private static final String KEY_CONTENT             = "content";
    private static final String KEY_CONTENT_TYPE        = "contentType";
    private static final String KEY_CONTENT_DISPOSITION = "disposition";
    private static final String KEY_DATA_HANDLER        = "dataHandler";


    //----------------------------------------------------------- PUBLIC METHODS
    
    /**
     * returns only header
     *
     * @param session opened mail server session
     * @param FID     parent folder id
     * @param FMID    mail id in the folder
     * @param msg     Message
     * @param loc     Locale
     * @return ItemMessage
     * @throws EntityException
     */
    public ItemMessage getPart_H(Session session,
                                 String FID,
                                 String FMID,
                                 Message msg,
                                 Locale loc)

      throws EntityException {


        if (log.isLoggable(Level.FINEST)) {
            log.finest(" start getPart_H ");
        }

        ItemMessage im;

        try {

            Message newMsg = new MimeMessage(session);

            // header
            String mid = Utility.setNewHeader(newMsg, msg);
            newMsg.setHeader(Def.HEADER_LABEL_MESSAGE_ID, mid);

            // default empty content
            newMsg.setContent(Def.DEFAULT_CONTENT_VALUE_BODY, 
                              Def.DEFAULT_CONTENT_VALUE_CONTENTTYPE);

            newMsg.saveChanges();
            
            String encodingType = getEncodingType(msg);

            im = setItemMessage(FMID, newMsg, FID, loc, encodingType);

        } catch (MessagingException me) {
            throw new EntityException(me);
        }

        return im;

    }


    /**
     *
     * returns header and body with byte quatity limitation
     *
     *
     * @param session  opened mail server session
     * @param FID      parent folder id
     * @param FMID     mail id in the folder
     * @param msg      Message
     * @param loc      Locale
     * @param numBytes int
     * @return ItemMessage
     * @throws EntityException
     */
    public ItemMessage getPart_HB(Session session,
                                  String FID,
                                  String FMID,
                                  Message msg,
                                  Locale loc,
                                  int numBytes)
      throws EntityException {

        ItemMessage im;
    
        try {
            
            Message newMsg = new MimeMessage(session);

            // header
            String mid = Utility.setNewHeader(newMsg, msg);
            newMsg.setHeader(Def.HEADER_LABEL_MESSAGE_ID, mid);

            // content
            Object    content    = msg.getDataHandler().getContent();
            ArrayList newContent = new ArrayList();
            
            if (content instanceof Multipart) {
                if (numBytes == -1){
                   newContent = getMultipart((Multipart) content, -1, true);
                } else {
                   newContent = getMultipart((Multipart) content, numBytes, true);
                }
            } else {
                if (numBytes == -1){
                   newContent = getSinglePart(msg, -1, true);
                } else {
                   newContent = getSinglePart(msg, numBytes, true);
                }   
            }

            InternalPart v = (InternalPart)newContent.get(0);
            
            newMsg.setContent(v.getPartValue(), v.getContentType());

            newMsg.saveChanges();
            
            String encodingType = getEncodingType(msg);

            im = setItemMessage(FMID, newMsg, FID, loc, encodingType);

        } catch (MessagingException me) {
            throw new EntityException(me);
        } catch (IOException ioe) {
            throw new EntityException(ioe);
        }

        return im;

    }



    /**
     *
     * if numBytes = -1 it returns all parts of the message 
     * (header, body, attachements) without byte quatity limitation.
     * otherwise it returns all parts of the message 
     * (header, body, attachements) but there is the byte number limitation
     *
     * @param session  opened mail server session
     * @param FID      parent folder id
     * @param FMID     mail id in the folder
     * @param msg      Message
     * @param loc      Locale
     * @param numBytes int
     * @return ItemMessage
     * @throws EntityException
     */
    public ItemMessage getPart_HBA(Session session,
                                   String FID,
                                   String FMID,
                                   Message msg,
                                   Locale loc,
                                   int numBytes)
      throws EntityException {

        ItemMessage im;
            
        try {
            
            if (numBytes == -1){
                
                String encodingType = getEncodingType(msg);
               
                im = setItemMessage(FMID, msg, FID, loc, encodingType);
               
            } else {
                
                Message newMsg = new MimeMessage(session);

                // header
                String mid = Utility.setNewHeader(newMsg, msg);
                newMsg.setHeader(Def.HEADER_LABEL_MESSAGE_ID, mid);

                // content
                Object content       = msg.getDataHandler().getContent();
                ArrayList newContent = new ArrayList();
                
                if (content instanceof Multipart) {
                    newContent = getMultipart((Multipart) content, numBytes, false);
                    // set the new content
                    MimeMultipart multipart = new MimeMultipart();
                    BodyPart part;
                    for (int i = 0; i < newContent.size(); i++) {
                        InternalPart v = (InternalPart)newContent.get(i);
                        part = new MimeBodyPart();
                        part.setContent(v.getPartValue(), v.getContentType());
                        if (v.getDisposition() != null){
                           part.setDisposition(v.getDisposition());
                        }
                        if (v.getFileName() != null){
                           part.setFileName(v.getFileName());
                        }
                        multipart.addBodyPart(part);
                    }
                    newMsg.setContent(multipart);
                } else {
                    newContent = getSinglePart(msg, numBytes, false);
                    // set the new content
                    InternalPart v = (InternalPart)newContent.get(0);
                    newMsg.setContent(v.getPartValue(), v.getContentType());
                }

                newMsg.saveChanges();

                String encodingType = getEncodingType(msg);

                im = setItemMessage(FMID, newMsg, FID, loc, encodingType);
            
            }

        } catch (MessagingException me) {
            throw new EntityException(me);
        } catch (IOException ioe) {
            throw new EntityException(ioe);
        }

        return im;
    }

    /**
     *
     * check if message contains non-ASCII characters
     * and set the ccorrect encoding type
     *
     * @param msg Message 
     * @return encodingType String
     */
    public String getEncodingType(Message msg)  throws EntityException {
        
        String encodingType = null;
        boolean asciiOnly;
        
        try {
            asciiOnly = isASCIIOnly(msg);
        } catch (MessagingException e1) {
            throw new EntityException(e1);
        }

        if (asciiOnly){
            encodingType = null;
        } else {
            encodingType = Def.ENCODE_QUOTED_PRINTABLE;                
        }
        return encodingType;
    }    
    
    //---------------------------------------------------------- PRIVATE METHODS
    

    /**
     * in the InternalBody there are the following values
     * the body; the content type; the stop value; the body size
     *
     * @param multipart Multipart
     * @param numBytes  int
     * @return InternalBody
     * @throws EntityException
     */
    private ArrayList getMultipart(Multipart multipart, 
                                   int numBytes,
                                   boolean getOnlyBody)
      throws EntityException {
        
        ArrayList out    = new ArrayList();
        ArrayList single = new ArrayList();
        
        try {
            
            int partsCount = multipart.getCount();
            
            InternalPart ip = null;
            
            int partsize  = 0;
            int totalsize = 0;
            
            boolean isBody = false;
            
            // loop on the Parts but it gets only body
            for (int i = 0; i < partsCount; i++) {
                
                single = getSinglePart(multipart.getBodyPart(i), numBytes, getOnlyBody);                
                
                ip = (InternalPart)single.get(0);
                
                isBody = ip.isBody().equalsIgnoreCase("true");
                
                if (getOnlyBody && isBody){
                    out.add(ip);
                    break; // stop after the body; exclude the attachment
                } else {
                    totalsize = totalsize + multipart.getBodyPart(i).getSize();                                    
                    if (isBody){
                        out.add(ip);
                    } else {
                       if (totalsize < numBytes){
                           out.add(ip);
                       } else {
                           // remove size counter, maybe there is another part I can add 
                           totalsize = totalsize - multipart.getBodyPart(i).getSize();
                       }                                        
                    }   
                }
            }
                        
        } catch (MessagingException me) {
            throw new EntityException(me);
        }     
        
        return out;
        
    }


    /**
     *
     * returns the body of the mail without attachment.
     * the method doesn't include the Part if disposition=Part.ATTACHMENT 
     * or disposition=Part.INLINE
     * <p>
     * output: InternalBody
     * <p> [0] (Body)
     * - string if we have a text/plain
     * - html   if we have a text/html
     * - empty string otherwise
     * <p> [1] (String)
     * - related content type
     * <p> [2] (String)
     * - flag to stop the Multipart loop
     * <p> [3] (int)
     * - size of the body
     *
     * 
     * @todo - manage UTF7 :"UTF7" ; "UNICODE-1-1-UTF-7",
     * "csUnicode11UTF7", "UNICODE-2-0-UTF-7"
     *
     * @param part Part
     * @param numBytes int
     * @return InternalBody
     * @throws EntityException
     */
    private ArrayList getSinglePart(Part part, 
                                    int numBytes,
                                    boolean getOnlyBody)
      throws EntityException {

        ArrayList out   = new ArrayList();
        
        InternalPart ip = new InternalPart(Def.DEFAULT_CONTENT_VALUE_BODY,
                                           Def.DEFAULT_CONTENT_VALUE_CONTENTTYPE,
                                           null,
                                           null,
                                           null,
                                           "false",
                                           0);
        out.add(ip);

        try {
            
            String      contentType     = null;
            String      disposition     = part.getDisposition();
            String      contentTypeOrig = part.getContentType().toLowerCase();
            int         size            = part.getSize();
            DataHandler dh              = part.getDataHandler();

            if (contentTypeOrig.equalsIgnoreCase("message/delivery-status") ||
                contentTypeOrig.equalsIgnoreCase("message/rfc822")) {
                return out;
            }

            if (contentTypeOrig.length() >= 10) {
                int index   = contentTypeOrig.indexOf(";");
                contentType = contentTypeOrig.toLowerCase().substring(0, index);
            }

            ContentType ct = new ContentType(part.getContentType());            
            String specifiedCharset = ct.getParameter("charset");

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Content Type Orig : [" + contentTypeOrig + "]" +
                           "Content Type New : [" + contentType + "]" +
                           "Disposition : [" + disposition + "]" +
                           "specifiedCharset [" + specifiedCharset + "]" +
                           "Content Class [" + part.getClass() + "]");
            }

            if (part instanceof MimeBodyPart || part instanceof Message) {
                
               if (disposition == null) {
                    
                  if (contentType.equals("text/plain") || contentType.equals("text/html") ) {

                      if ("utf-7".equalsIgnoreCase(specifiedCharset) ||
                          "unicode-1-1-utf-7".equalsIgnoreCase(specifiedCharset)) {
                          ip.setPartValue(Def.DEFAULT_CONTENT_VALUE_BODY);
                          ip.setContentType(contentTypeOrig);
                          ip.setIsBody("true");
                          ip.setSize(0);  
                      } else {
                          ip.setPartValue(part.getContent());
                          ip.setContentType(contentTypeOrig);
                          ip.setIsBody("true");
                          ip.setSize(size); 
                          
                          // eventually trunc the body
                          if (numBytes != -1){
                             if (numBytes < ip.getSize()) {
                                // get truncated content
                                byte[] temp = ip.getPartValue().toString().getBytes();
                                ip.setPartValue(new String(temp, 0, numBytes));
                             } 
                          }                          
                      }
                      
                 }    
                  
                 if (contentType.equals("multipart/alternative")) {
                      MimeMultipart mmp = (MimeMultipart) part.getContent();
                      out = getMultipart(mmp, numBytes, getOnlyBody);
                 }
                  
              } else {
                  if (!getOnlyBody){
                      if (disposition.equalsIgnoreCase(Part.ATTACHMENT) || 
                          disposition.equalsIgnoreCase(Part.INLINE)) {
                           ip.setPartValue(part.getContent());
                           ip.setContentType(contentTypeOrig);
                           ip.setDisposition(disposition);
                           ip.setFileName(part.getFileName());
                           ip.setDataHandler(dh);
                      }            
                  }                       
              }                             
          }
            
        } catch (MessagingException me) {
            throw new EntityException(me);
        } catch (IOException ioe) {
            throw new EntityException(ioe);
        }

        return out;
    }
    
    
    /**
     * <p>Retrieve the text parts of a message or a multipart.<p>
     *
     * @param part the input object (message or multipart)
     * @return an array with the input object's text parts
     * @throws MessagingException
     */
    private static Part[] getTextParts(Object part)  throws MessagingException {
        List partsList;
        String contentType;
        int partsCount;
        Multipart multipart;
        Part[] parts;

        if (part == null) {
            return null;
        }

        partsList = new LinkedList();
        try {
            if (part instanceof Message && 
                ((Message) part).getContent() instanceof Multipart) {
                return getTextParts(((Message) part).getContent());
            }
        }
        catch (IOException e) {
            throw new MessagingException(e.getMessage(), e);
        }

        if (part instanceof Multipart) {
            multipart = (Multipart) part;
            partsCount = multipart.getCount();
            for (int i = 0; i < partsCount; i++) {
                parts = getTextParts(multipart.getBodyPart(i));
                for (int j = 0; j < parts.length; j++) {
                    partsList.add(parts[j]);
                }
            }
        } else {
            contentType = ((Part) part).getContentType();
            if (contentType.toLowerCase().startsWith("text/plain") ||
                contentType.toLowerCase().startsWith("text/html")) {
                partsList.add(part);
            }
        }

        return (Part[]) partsList.toArray(new Part[0]);
    }
    
    
    /**
     * <p>Checks if the given message has text parts formed only of ASCII characters</p>
     *
     * @param message the input message
     * @return <i>true</i> if message is ASCII-only, <i>false</i> otherwise
     * @throws MessagingException
     */
    private static boolean isASCIIOnly(Message message)
            throws MessagingException {
        Object rawContent;
        Part[] textParts;

        textParts = getTextParts(message);

        if (textParts == null) {
            throw new MessagingException("No text parts found in message.");
        }
        
        for (int i = 0; i < textParts.length; i++) {
            try {
                rawContent = textParts[i].getContent();
            } catch (IOException e) {
                throw new MessagingException(e.getMessage(), e);
            }

            if (rawContent == null) {
                continue;
            }
            
            if (!isASCIIOnly(rawContent.toString())) {
                return false;
            }
            
        }

        return true;
    }
    
    
    /**
     * <p>Checks if the given text is formed only of ASCII characters</p>
     *
     * @param text the input text
     * @return <i>true</i> if text is ASCII-only, <i>false</i> otherwise
     */
    private static boolean isASCIIOnly(String text) {
        char[] chars;

        if (text == null) {
            return false;
        }

        chars = text.toCharArray();
        for (int i = 0; i < chars.length; i++) {
            if (chars[i] > 126) {
                return false;
            }
        }
        return true;
    }   
    
    
}
